import os
from copy import deepcopy
from datetime import datetime
from time import sleep
from traceback import format_exc
from typing import List, Dict

import schedule

from divinegift.config import Settings
from divinegift.logger import Logger
from divinegift.main import get_args, get_log_param, get_list_files
from divinegift.sender import Sender


class Application:
    def __init__(self):
        self.st = datetime.now()
        self.sp = datetime.now()
        self.args = get_args()
        self.lp = get_log_param(self.args)
        self.logger = Logger()
        self.logger.set_log_level(**self.lp)

        self.settings = Settings(logger_=self.logger)
        self.settings_edited_at = None
        self.conf_file_path = self.settings.find_config_file(self.args.get('c', 'settings.yaml'))
        self.cipher_key_path = self.args.get('ck', 'key.ck')

        self.print_intro()

        self.logger.log_info(f'{"=" * 20} START {"=" * 20}')

    def __repr__(self):
        return f'Application()'

    def update_settings(self, encrypt_config=True, log_changes=False, use_yaml=True, encoding='utf-8'):
        """
        Update settings at service if it changes
        :param encrypt_config: if True will encrypt key fields in self.init_config method
        :param log_changes: if True will show changes in logs
        :param use_yaml: if True will use yaml format instead json
        :param encoding: encoding
        :return:
        """
        settings_edited_at = os.path.getmtime(self.conf_file_path)
        if settings_edited_at != self.settings_edited_at:
            self.set_settings(encrypt_config, log_changes, use_yaml, encoding)

    def run(self):
        pass

    def set_run_schedule(self):
        """
        Set schedule to run service every n time_units.
        Should be set settings variable `service_run_schedule` like string:
        every n seconds/minutes/hours/days/weeks [at [[00:]00]:00].
        E.g 1: service_run_schedule: every 10 seconds
        E.g 2: service_run_schedule: every 1 minutes at :00
        E.g 3: service_run_schedule every 1 day at 10:30:00
        :return:
        """
        service_run_schedule = self.get_settings('service_run_schedule', '').split()
        if service_run_schedule:
            every = service_run_schedule[1]
            time_unit = service_run_schedule[2]
            try:
                at = service_run_schedule[4]
                at_str = f'.at("{at}")'
            except IndexError:
                at_str = ''
            eval(f"schedule.every({every}).{time_unit}{at_str}.do(self.run)")
        else:
            schedule.every(5).seconds.do(self.run)

    def run_service(self):
        self.set_run_schedule()
        while True:
            schedule.run_pending()
            sleep(1)

    def set_settings(self, encrypt_config=True, log_changes=False, use_yaml=True, encoding='utf-8'):
        self.settings.parse_settings(self.conf_file_path, self.cipher_key_path,
                                     encoding=encoding, log_changes=log_changes, use_yaml=use_yaml)
        self.settings_edited_at = os.path.getmtime(self.conf_file_path)

        if encrypt_config:
            self.init_config()

    def init_config(self):
        """
        Encrypt config if not exist key.ck file at self.cipher_key_path
        Override it if you have another passwords which should be encrypted
        :return:
        """
        if not os.path.exists(self.cipher_key_path):
            self.settings.initialize_cipher(ck_fname=self.cipher_key_path)
            self.encrypt_password('email_conf', 'pwd')           # Change it for your config
            self.settings.save_settings(self.conf_file_path)

    def encrypt_password(self, connection_name: str, pass_field: str = 'db_pass'):
        try:
            self.settings.encrypt_password(connection_name, pass_field)
        except Exception as ex:
            self.logger.log_err(f'Could not encrypt password: {ex}')

    def get_settings(self, param=None, default=None):
        """
        Return value from settings
        :param param: desired settings key
        :param default: if param not exist return this
        :return:
        """
        return self.settings.get_settings(param, default)

    def print_intro(self):
        print(f'Process {self.args.get("name")} started!')
        log_place = os.path.join(self.lp.get("log_dir"), self.lp.get("log_name")) if self.lp.get(
            "log_name") else "ON SCREEN"
        print(f'Log will be here: {log_place}')

    def log_debug(self, *args, separator: str = ' '):
        self.logger.log_debug(*args, separator)

    def log_info(self, *args, separator: str = ' '):
        self.logger.log_info(*args, separator)

    def log_warning(self, *args, separator: str = ' '):
        self.logger.log_warn(*args, separator)

    def log_warn(self, *args, separator: str = ' '):
        self.logger.log_warn(*args, separator)

    def log_err(self, *args, separator: str = ' ', src: str = None, mode: List = None, channel: Dict = None):
        self.logger.log_err(*args, separator, src, mode, channel)

    def log_crit(self, *args, separator: str = ' '):
        self.logger.log_crit(*args, separator)

    def log_err_notif(self, *args, separator: str = ' '):
        mode = self.get_settings('monitoring')
        channel = deepcopy(self.get_settings('monitoring_channel'))
        if 'email' in mode or 'email_attach' in mode:
            channel['email'] = deepcopy(self.get_settings('email_conf'))
            channel['email']['TO'] = channel.pop('email_to')
        self.log_err(*args, separator=separator)
        src = self.get_settings('service_name', self.args.get('name'))
        self.notify(*args, separator=separator, src=src, mode=mode, channel=channel)

    def notify(self, *args, separator: str = ' ', src: str = None, mode: List = None, channel: Dict = None):
        """
        :param args:
        :param separator:
        :param msg: Error message which will logged
        :param src: Source which raised error
        :type src: string
        :param mode: List of monitoring's mode (telegram, email, email_attach, slack)
        :type mode: list
        :param channel: Dict with parameters of monitoring (e.g. {'telegram': -1001343660695}
        :type channel: dict
        :return:
        """
        error_txt = f'''An error has occurred in {src}
        Error text: {separator.join([str(x) for x in args])}\n{format_exc()}'''

        sender = Sender()

        if mode:
            if 'telegram' in mode:
                sender.send_telegram(error_txt, chat_id=channel.get('telegram', -1001343660695))
            if 'slack' in mode:
                sender.send_slack(error_txt)
            if 'email' in mode:
                sender.send_mail(error_txt, f'{src} ERROR', **channel.get('email'))
            if 'email_attach' in mode:
                if self.lp.get('log_dir', '.') and self.lp.get('log_name'):
                    log = get_list_files(self.lp.get('log_dir'), self.lp.get('log_name'), add_path=True)
                else:
                    log = []
                sender.send_mail(error_txt, f'{src} ERROR', attachments=log, **channel.get('email_attach'))



from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from typing import Optional

from divinegift import db
from divinegift.errors import DBNotConnectedError


class AimsUtils:
    def __init__(self):
        self.connection: Optional[db.DBClient] = None

    def connect(self, db_conn):
        self.connection = db.DBClient(db_conn)

    def check_connection(self):
        if not self.connection:
            raise DBNotConnectedError(f'Use connect() first!')

    def get_task_code(self, index, rectype, table_scheme):
        """
        Определение задачи по IND (7.4.3 Crew Administration and Training Qualification Codes)
        :param index: Индекс из БД
        :param rectype: 0 - Получить название задачи; 1 - Код задачи тренера; 2 - Код задачи тренируемого
        :param table_scheme: Схема БД
        :return:
        """
        self.check_connection()

        nIND = None
        if 1 <= index <= 41:
            nIND = index
        elif 1 + 127 <= index <= 41 + 127:
            nIND = index - 127

        if nIND:
            d = self.connection.get_data_row(f"""select * from {table_scheme}.TABLES vt
            where vt.RECDESC = 'CREWCODES2'""")
            fields = d.get('fields')

            trainee_code = ''.join([c.decode() if isinstance(c, bytes) else chr(c) for c in fields[nIND + 30 - 1:nIND + 30]])
            trainer_code = ''.join([c.decode() if isinstance(c, bytes) else chr(c) for c in fields[3059 + nIND * 71 - 1:3059 + nIND * 71]])
            task_name = ''.join([c.decode() if isinstance(c, bytes) else chr(c) for c in fields[3000 + nIND * 71 - 1:3000 + nIND * 71 + 58] if c not in (0, 1)])

            if rectype == 0:
                return task_name
            elif rectype == 1:
                return trainer_code
            elif rectype == 2:
                return trainee_code

    @staticmethod
    def dst_rule_to_date(year, month, week_day, week_day_num, minutes):
        """
        Определение даты перевода летнего/зимнего времени (для AIMS-овского формата представления правила)
        :param year: год, для которого необходимо определить дату
        :param month: месяц
        :param week_day: день недели (понедельник - 1, вторник - 2, и тд)
        :param week_day_num: порядковый номер дня недели в месяце (0 - последний, 1 - первый, 2 - второй, и тд)
        :param minutes: количество минут, во сколько осуществляется перевод
        :return: datetime
        """
        step = 0
        first_day = datetime(year, month, 1)
        init_day = first_day - timedelta(days=first_day.weekday()) + week_day - 1

        if init_day < first_day:
            init_day += 7

        res = datetime(1980, 1, 1)
        while init_day < first_day + relativedelta(months=1, days=-1):
            step += 1
            if week_day_num == 0:
                res = init_day
            elif step == week_day_num:
                res = init_day
                break
            init_day += 7
        return res + timedelta(days=minutes // 60 // 24)


if __name__ == '__main__':
    pass



from string import Template
import os
from time import sleep
from typing import Union, Optional, Tuple, Any, List, Dict

from deprecation import deprecated

from divinegift.errors import MethodNotAllowedError

try:
    from sqlalchemy import create_engine, MetaData, Table, text
    from sqlalchemy.engine import Engine
    from sqlalchemy.exc import DatabaseError, ResourceClosedError
    from sqlalchemy.engine.base import Connection as SQLAlchemyConnection
except ImportError:
    raise ImportError("sqlalchemy isn't installed. Run: pip install -U sqlalchemy")

from divinegift import version


class DBClient:
    def __init__(self, db_conn: dict, future=True, do_initialize=True, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.db_conn = db_conn
        self.future = future

        self.engine: Optional[Engine] = None
        self.conn: Optional[SQLAlchemyConnection] = None
        self.metadata: Optional[MetaData] = None

        if do_initialize:
            self.create_engine()

    def get_conn_str(self) -> str:
        if self.db_conn.get('dialect') == 'mssql+pytds':
            from sqlalchemy.dialects import registry
            registry.register("mssql.pytds", "sqlalchemy_pytds.dialect", "MSDialect_pytds")
        if self.db_conn.get('db_host') and self.db_conn.get('db_port'):
            if 'oracle' in self.db_conn.get('dialect').lower() and '.orcl' in self.db_conn.get('db_name'):
                connect_str = '{dialect}://{db_user}:{db_pass}@{db_host}:{db_port}/?service_name={db_name}'.format(
                    **self.db_conn)
            else:
                connect_str = '{dialect}://{db_user}:{db_pass}@{db_host}:{db_port}/{db_name}'.format(**self.db_conn)
        else:
            connect_str = '{dialect}://{db_user}:{db_pass}@{db_name}'.format(**self.db_conn)

        return connect_str

    def create_engine(self) -> None:
        connect_str = self.get_conn_str()
        self.engine = create_engine(connect_str, future=self.future, *self.args, **self.kwargs)

    def create_conn(self) -> None:
        if not self.conn:
            self.conn = self.engine.connect()

    def create_raw_conn(self) -> None:
        if not self.conn:
            self.conn = self.engine.raw_connection()

    def create_metadata(self) -> None:
        if not self.metadata:
            self.create_conn()
            self.metadata = MetaData(bind=self.engine)

    def set_args(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def set_conn(self) -> None:
        """
        Create connection for SQLAlchemy
        :return: Engine, Connection, Metadata
        """
        self.create_engine()
        self.create_conn()
        self.create_metadata()

    def set_raw_conn(self) -> None:
        """
        Create raw connection for SQLAlchemy
        :return: Connection
        """
        connect_str = self.get_conn_str()

        self.engine = create_engine(connect_str, *self.args, **self.kwargs)
        self.conn = self.engine.raw_connection()
        self.metadata = MetaData(bind=self.conn)

    def get_conn(self, fields: Union[str, list] = 'conn') -> Tuple[Optional[Any], ...]:
        """
        Return connection fields: engine, conn, metadata
        :param fields: str or list of str
        :return: tuple or single value
        """
        if isinstance(fields, str):
            return self.__dict__.get(fields)
        elif isinstance(fields, list):
            return tuple([self.__dict__.get(x) for x in fields])
        else:
            raise MethodNotAllowedError('This type of fields is not allowed')

    def close_conn(self):
        self.conn.close()
        self.engine.dispose()
        self.conn = None

    def check_connection_status(self):
        if 'oracle' in self.db_conn.get('dialect').lower():
            d = self.get_data_row('select dummy from dual')
        else:
            d = self.get_data_row("select 'x' as dummy")
        if (v := d.get('dummy')) is None or v != 'x':
            self.close_conn()
            self.set_conn()

    @staticmethod
    def get_sql(filename: str, encoding: str = 'utf-8') -> str:
        """
        Get sql string from file
        :param filename: File name
        :param encoding: Encoding of file
        :return: String with sql
        """
        with open(filename, 'r', encoding=encoding) as file:
            sql = file.read()
        return sql

    def get_data(self, sql: str, encoding: str = 'utf-8', print_script=False, max_attempts=5, **kwargs) -> List[Dict]:
        """
        Get raw data from sql data as dict
        :param sql: File with sql which need to execute or script itself
        :param encoding: Encoding of file
        :param print_script: Print script text to console
        :param max_attempts: Max count  of tries to run script
        :param kwargs: List with additional data
        :return: Dictionary
        """
        if not self.conn:
            self.create_conn()

        if os.path.exists(sql):
            script_t = Template(self.get_sql(sql, encoding))
        else:
            script_t = Template(sql)
        script = script_t.safe_substitute(**kwargs)

        if print_script:
            print(script)

        res = self._execute(script, max_attempts)

        return res

    def get_data_row(self, sql: str, index: int = 0, encoding: str = 'utf-8',
                     print_script=False, max_attempts=5, **kwargs) -> Optional[Dict]:
        """
        Get raw data from sql data as dict
        :param sql: File with sql which need to execute or script itself
        :param index: index of returning row
        :param encoding: Encoding of file
        :param print_script: Print script text to console
        :param max_attempts: Max count  of tries to run script
        :param kwargs: List with additional data
        :return: Dictionary
        """
        if not self.conn:
            self.create_conn()

        if os.path.exists(sql):
            script_t = Template(self.get_sql(sql, encoding))
        else:
            script_t = Template(sql)
        script = script_t.safe_substitute(**kwargs)

        if print_script:
            print(script)

        res = self._execute(script, max_attempts)

        try:
            res = res[index]
        except:
            res = None

        return res

    def _execute(self, script, max_attempts):
        res = []
        trns = None
        for _ in range(max_attempts):
            try:
                if not self.future:
                    trns = self.conn.begin()
                res = self.conn.execute(text(script))
                try:
                    res = [dict(row) for row in res]
                except ResourceClosedError:
                    pass
                self.commit(trns)
                break
            except DatabaseError:
                self.rollback(trns)
                try:
                    self.close_conn()
                except:
                    pass
                sleep(10)
                self.set_conn()
        return res

    def run_script(self, sql: str, encoding: str = 'utf-8', print_script=False, max_attempts=5, **kwargs) -> None:
        """
        Run custom script
        :param sql: File with sql which need to execute
        :param encoding: Encoding of file
        :param print_script: Print script text to console
        :param max_attempts: Max count  of tries to run script
        :param kwargs: List with additional data
        :return: None
        """
        if not self.conn:
            self.create_conn()

        if os.path.exists(sql):
            script_t = Template(self.get_sql(sql, encoding))
        else:
            script_t = Template(sql)
        script = script_t.safe_substitute(**kwargs)

        if print_script:
            print(script)

        self._execute(script, max_attempts)

    def commit(self, transaction=None):
        if transaction:
            transaction.commit()
        else:
            self.conn.commit()

    def rollback(self, transaction=None):
        if transaction:
            transaction.rollback()
        else:
            self.conn.rollback()

    def begin_transaction(self):
        return self.engine.begin()


@deprecated(deprecated_in='1.3.9', current_version=version, details='Use class Connection instead')
def get_conn(db_conn: dict):
    conn_obj = DBClient(db_conn)
    conn_obj.set_conn()
    return conn_obj.get_conn(['engine', 'conn', 'metadata'])


@deprecated(deprecated_in='1.3.9', current_version=version, details='Use class Connection instead')
def get_raw_conn(db_conn: dict):
    conn_obj = DBClient(db_conn)
    conn_obj.set_raw_conn()
    return conn_obj.get_conn('conn')


@deprecated(deprecated_in='1.3.9', current_version=version, details='Use class Connection instead')
def get_sql(filename: str, encoding: str = 'utf-8'):
    """
    Get sql string from file
    :param filename: File name
    :param encoding: Encoding of file
    :return: String with sql
    """
    file = open(filename, 'r', encoding=encoding)
    sql = file.read()
    file.close()
    return sql


@deprecated(deprecated_in='1.3.9', current_version=version, details='Use class Connection instead')
def get_data(sql: str, db_conn, encoding: str = 'utf-8', print_script=False, **kwargs):
    conn_obj = DBClient(db_conn)
    if isinstance(db_conn, dict):
        conn_obj.set_conn()
    else:
        conn_obj.conn = db_conn
    ress = conn_obj.get_data(sql, encoding, print_script, **kwargs)

    return ress


@deprecated(deprecated_in='1.3.9', current_version=version, details='Use class Connection instead')
def get_data_row(sql: str, db_conn: dict, index: int = 0, encoding: str = 'utf-8', **kwargs):
    conn_obj = DBClient(db_conn)
    if isinstance(db_conn, dict):
        conn_obj.set_conn()
    else:
        conn_obj.conn = db_conn
    ress = conn_obj.get_data_row(sql, index, encoding, **kwargs)

    return ress


@deprecated(deprecated_in='1.3.9', current_version=version, details='Use class Connection instead')
def run_script(sql: str, db_conn: dict, encoding: str = 'utf-8', **kwargs):
    conn_obj = DBClient(db_conn)
    if isinstance(db_conn, dict):
        conn_obj.set_conn()
    else:
        conn_obj.conn = db_conn
    conn_obj.run_script(sql, encoding, **kwargs)


Connection = DBClient


if __name__ == '__main__':
    pass



from string import printable
from typing import Union, Optional, List, Dict, Iterable

from deprecation import deprecated

from divinegift import main, version

import os
from datetime import datetime, timedelta, date

try:
    import openpyxl
    from openpyxl.utils.exceptions import InvalidFileException, IllegalCharacterError
    from openpyxl.styles import Border, Side, Alignment, Font
except ImportError:
    raise ImportError("openpyxl isn't installed. Run: pip install -U openpyxl")
try:
    import xlrd
except ImportError:
    pass
try:
    import xlsxwriter
except ImportError:
    raise ImportError("xlsxwriter isn't installed. Run: pip install -U xlsxwriter")
try:
    import xlwt
except ImportError:
    raise ImportError("xlwt isn't installed. Run: pip install -U xlwt")

BORDER_NONE = {
    'left': {
        'style': None,
        'color': 'FFFFFFFF'
    },
    'right': {
        'style': None,
        'color': 'FFFFFFFF'
    },
    'bottom': {
        'style': None,
        'color': 'FFFFFFFF'
    },
    'top': {
        'style': None,
        'color': 'FFFFFFFF'
    },
}

BORDER_BLACK_LINE = {
    'style': 'thin',
    'color': '000000'
}

BORDER = {
    'none': BORDER_NONE,
    'all': {
        'left': BORDER_BLACK_LINE,
        'right': BORDER_BLACK_LINE,
        'bottom': BORDER_BLACK_LINE,
        'top': BORDER_BLACK_LINE,
    },
    'left': {
        **BORDER_NONE,
        'left': BORDER_BLACK_LINE,
    },
    'right': {
        **BORDER_NONE,
        'right': BORDER_BLACK_LINE,
    },
    'bottom': {
        **BORDER_NONE,
        'bottom': BORDER_BLACK_LINE,
    },
    'top': {
        **BORDER_NONE,
        'top': BORDER_BLACK_LINE,
    }
}

FONT = {
    'default': {
        'name': 'Calibri',
        'size': 11,
        'bold': False,
        'italic': False,
        'underline': 'none',
        'strike': False,
        'color': '000000'
    }
}

NUMBER_FORMAT = {
    'date': {'number_format': 'dd.mm.yyyy'},
    'datetime': {'number_format': 'dd.mm.yyyy hh:mm:ss'},
    'other': {'number_format': 'general'}
}


class Excel:
    def __init__(self, filename: str = None):
        self.filename = filename
        self.styles = {}

        self.workbook = None
        self.worksheets = {}
        self.current_worksheet = None

        self.rows_qty: int = 0

    def get_file(self):
        raise NotImplementedError(f'Method get_file not implemented in class {self.__class__.__name__}')

    def create_file(self):
        raise NotImplementedError(f'Method create_file not implemented in class {self.__class__.__name__}')

    def open_file(self, filename: str = None):
        # Если указанной папки не существует - создаем ее (включая родительские, если нужно)
        if filename:
            self.filename = filename
        try:
            fd, fn = os.path.split(self.filename)
            if fd:
                main.check_folder_exist(fd)
        except TypeError:
            pass

    def save_file(self):
        raise NotImplementedError(f'Method save_name not implemented in class {self.__class__.__name__}')

    def add_worksheet(self, worksheet: str):
        raise NotImplementedError(f'Method add_worksheet not implemented in class {self.__class__.__name__}')

    def rename_worksheet(self, worksheet_old: str, worksheet_new: str):
        raise NotImplementedError(f'Method rename_worksheet not implemented in class {self.__class__.__name__}')

    def select_worksheet(self, worksheet):
        self.current_worksheet = self.worksheets.get(worksheet)

    def add_cell(self, row: int, col: int, value: Union[None, str, float, int, date, datetime],
                 style: Optional[Dict] = None):
        raise NotImplementedError(f'Method add_cell not implemented in class {self.__class__.__name__}')

    def add_cell_coords(self, coords: str, value: Union[None, str, float, int, date, datetime],
                        style: Optional[Dict] = None):
        raise NotImplementedError(f'Method add_cell_coords not implemented in class {self.__class__.__name__}')

    def add_row(self, row: int, list_cells: Iterable[Union[None, str, float, int, date, datetime]],
                start_col: int = 0, styles: Union[None, List, Dict] = None):
        raise NotImplementedError(f'Method add_row not implemented in class {self.__class__.__name__}')

    def add_column(self, col: int, list_cells: Iterable[Union[None, str, float, int, date, datetime]],
                   start_row: int = 1, styles: Union[None, List, Dict] = None):
        raise NotImplementedError(f'Method add_column not implemented in class {self.__class__.__name__}')

    def add_merge_cell(self, first_row: int, first_col: int, last_row: int, last_col: int,
                       value: Union[None, str, float, int, date, datetime], style: Optional[Dict] = None):
        raise NotImplementedError(f'Method add_merge_row not implemented in class {self.__class__.__name__}')

    def add_merge_cell_coords(self, first_coords: str, last_coords,
                              value: Union[None, str, float, int, date, datetime], style: Optional[Dict] = None):
        raise NotImplementedError(f'Method add_merge_cell_coords not implemented in class {self.__class__.__name__}')

    def add_list_rows(self, first_row: int, first_col, list_rows: List, styles: Union[None, List, Dict] = None):
        raise NotImplementedError(f'Method add_list_rows not implemented in class {self.__class__.__name__}')

    def set_cell_dimension(self, row: int, col: int, height: int = 15, width: int = 10):
        raise NotImplementedError(f'Method set_cell_dimension not implemented in class {self.__class__.__name__}')

    def set_columns_width(self, list_columns_width: List[int], start_col: int = 1):
        raise NotImplementedError(f'Method set_columns_width not implemented in class {self.__class__.__name__}')

    def set_rows_height(self, list_rows_height: List[int], start_row: int = 1):
        raise NotImplementedError(f'Method set_rows_height not implemented in class {self.__class__.__name__}')

    def create_simple_excel(self, filename, data: List[Union[Dict, List]], column_width: Optional[List] = None,
                            column_names: Optional[List] = None, styles: Union[None, List, Dict] = None):
        raise NotImplementedError(f'Method create_simple_excel not implemented in class {self.__class__.__name__}')

    def get_max_row(self):
        raise NotImplementedError(f'Method get_max_row not implemented in class {self.__class__.__name__}')

    def get_max_column(self):
        raise NotImplementedError(f'Method get_max_column not implemented in class {self.__class__.__name__}')

    def read_file(self, excel_header: Union[list, dict], sheet_name: str = None,
               int_columns: list = None, date_columns: list = None, start_row: int = 1):
        raise NotImplementedError(f'Method read_file not implemented in class {self.__class__.__name__}')


class ExcelX(Excel):
    def open_file(self, filename: str = None, rewrite=True):
        if filename and not filename.endswith('xlsx'):
            filename += '.xlsx'

        super(ExcelX, self).open_file(filename=filename)

        if os.path.exists(self.filename) and not rewrite:
            self.get_file()
        else:
            self.create_file()

        self.worksheets = {w.title: w for w in self.workbook.worksheets}

    def get_file(self):
        self.workbook = openpyxl.load_workbook(self.filename)

    def create_file(self):
        self.workbook = openpyxl.workbook.Workbook()

    def add_worksheet(self, worksheet: str):
        self.worksheets[worksheet] = self.workbook.create_sheet(worksheet)

    def rename_worksheet(self, worksheet_old: str, worksheet_new: str):
        ws = self.workbook[worksheet_old]
        ws.title = worksheet_new
        self.worksheets = {w.title: w for w in self.workbook.worksheets}
        self.select_worksheet(worksheet_new)

    def select_worksheet(self, worksheet):
        if not self.worksheets.get(worksheet):
            self.add_worksheet(worksheet)
        self.current_worksheet = self.worksheets.get(worksheet)

    def save_file(self, new_filename: str = None):
        filename = self.filename if not new_filename else new_filename
        self.workbook.save(filename)

    def add_cell_coords(self, coords: str, value: Union[None, str, float, int, date, datetime],
                        style: Optional[Dict] = None):
        if not self.current_worksheet:
            self.current_worksheet = self.workbook.active
        cell = self.current_worksheet[coords]
        self.add_cell(cell.row, cell.column, value, style)

    def add_cell(self, row: int, col: int, value: Union[None, str, float, int, date, datetime],
                 style: Optional[Dict] = None):
        if not self.current_worksheet:
            self.current_worksheet = self.workbook.active
        try:
            cell = self.current_worksheet.cell(row, col, value)
        except IllegalCharacterError:
            value = ''.join(c for c in value if c in printable)
            cell = self.current_worksheet.cell(row, col, value)
        if style:
            if style.get('border'):
                b = style.get('border')
                left = b.get('left', {})
                right = b.get('right', {})
                bottom = b.get('bottom', {})
                top = b.get('top', {})
                cell.border = Border(left=Side(style=left.get('style', None), color=left.get('color', 'FF000000')),
                                     right=Side(style=right.get('style', None), color=right.get('color', 'FF000000')),
                                     bottom=Side(style=bottom.get('style', None),
                                                 color=bottom.get('color', 'FF000000')),
                                     top=Side(style=top.get('style', None), color=top.get('color', 'FF000000')))
            if style.get('bold'):
                cell.font = Font(bold=True)
            if style.get('italic'):
                cell.font = Font(italic=True)
            if style.get('font'):
                f = style.get('font')
                cell.font = Font(name=f.get('font', 'Calibri'),
                                 size=f.get('size', 11),
                                 bold=f.get('bold', False),
                                 italic=f.get('italic', False),
                                 underline=f.get('underline', 'none'),
                                 strike=f.get('strike', False),
                                 color=f.get('color', '000000'))
            if style.get('alignment'):
                cell.alignment = Alignment(horizontal=style.get('alignment').get('horizontal', 'general'),
                                           vertical=style.get('alignment').get('vertical', 'bottom'),
                                           wrap_text=style.get('alignment').get('wrap_text', None))
            if style.get('number_format'):
                cell.number_format = style.get('number_format', 'general')

    def add_row(self, row: int, list_cells: Iterable[Union[None, str, float, int, date, datetime]],
                start_col: int = 1, styles: Union[None, List, Dict] = None):
        if isinstance(list_cells, dict):
            list_cells_ = list(list_cells.values())
        else:
            list_cells_ = list_cells
        for col, cell in enumerate(list_cells_):
            style = None
            if styles:
                if isinstance(styles, list):
                    style = styles[col]
                elif isinstance(styles, dict):
                    style = styles
            self.add_cell(row, start_col + col, cell, style)
        self.rows_qty += 1

    def add_column(self, col: int, list_cells: Iterable[Union[None, str, float, int, date, datetime]],
                   start_row: int = 1, styles: Union[None, List, Dict] = None):
        if isinstance(list_cells, dict):
            list_cells_ = list(list_cells.values())
        else:
            list_cells_ = list_cells
        for row, cell in enumerate(list_cells_):
            style = None
            if styles:
                if isinstance(styles, list):
                    style = styles[row]
                elif isinstance(styles, dict):
                    style = styles
            self.add_cell(start_row + row, col, cell, style)
            self.rows_qty += 1

    def add_list_rows(self, first_row: int, first_col: int, list_rows: Iterable[Union[Dict, List]],
                      styles: Union[None, List, Dict] = None):
        rows_cnt = len(list_rows)
        if styles:
            styles_cnt = len(styles)
        else:
            styles_cnt = 0
        for row_index, row_ in enumerate(list_rows):
            if isinstance(row_, dict):
                row = list(row_.values())
            else:
                row = row_
            if isinstance(styles, list) and styles_cnt == rows_cnt:
                self.add_row(first_row + row_index, row, first_col, styles[row_index])
            else:
                self.add_row(first_row + row_index, row, first_col, styles)

    def add_merge_cell(self, first_row: int, first_col: int, last_row: int, last_col: int,
                       value: Union[None, str, float, int, date, datetime], style: Optional[Dict] = None):
        self.add_cell(first_row, first_col, value, style)
        self.current_worksheet.merge_cells(start_row=first_row, start_column=first_col,
                                           end_row=last_row, end_column=last_col)

    def add_merge_cell_coords(self, first_coords: str, last_coords,
                              value: Union[None, str, float, int, date, datetime], style: Optional[Dict] = None):
        if not self.current_worksheet:
            self.current_worksheet = self.workbook.active
        first_cell = self.current_worksheet[first_coords]
        last_cell = self.current_worksheet[last_coords]
        self.add_merge_cell(first_cell.row, first_cell.column, last_cell.row, last_cell.column, value, style)

    def set_cell_dimension(self, row: int, col: int, height: Union[int, float] = 15, width: Union[int, float] = 10):
        cell = self.current_worksheet.cell(row, col)
        self.current_worksheet.column_dimensions[cell.column_letter].width = width
        self.current_worksheet.row_dimensions[row].height = height

    def set_rows_height(self, list_rows_height: Iterable[float], start_row: int = 1):
        for row, height in enumerate(list_rows_height):
            # cell = self.current_worksheet.cell(row + 1, 0)
            self.current_worksheet.row_dimensions[start_row + row].height = height

    def set_columns_width(self, list_columns_width: Iterable[float], start_col: int = 1):
        for col, width in enumerate(list_columns_width):
            cell = self.current_worksheet.cell(1, start_col + col)
            self.current_worksheet.column_dimensions[cell.column_letter].width = width

    def create_simple_excel(self, filename, data: Iterable[Union[Dict, List]], column_width: Optional[Iterable] = None,
                            column_names: Optional[Iterable] = None, styles: Union[None, List, Dict] = None,
                            worksheet: str = 'Sheet', rewrite: bool = True, start_row: int = 1):
        header_style = {'bold': True, 'border': BORDER.get('bottom'),
                        'alignment': {'horizontal': 'center', 'wrap_text': True}}

        self.open_file(filename, rewrite)
        if rewrite:
            self.rename_worksheet('Sheet', worksheet)
        self.select_worksheet(worksheet)

        if column_width:
            self.set_columns_width(column_width)

        _start_row = start_row

        if column_names:
            self.add_row(_start_row, column_names, styles=header_style)
            _start_row += 1

        if any(isinstance(e, dict) for e in data):
            self.add_list_rows(_start_row, 1, [list(x.values()) for x in data], styles)
        else:
            self.add_list_rows(_start_row, 1, data, styles)
        self.save_file()

    def get_max_row(self):
        if not self.current_worksheet:
            self.current_worksheet = self.workbook.active
        return self.current_worksheet.max_row

    def get_max_column(self):
        if not self.current_worksheet:
            self.current_worksheet = self.workbook.active
        return self.current_worksheet.max_colummn

    def read_file(self, excel_header: list, sheet_name: str = None,
               int_columns: list = None, date_columns: list = None, start_row: int = 1):
        excel_arr = []
        columns = excel_header

        ws = self.worksheets.get(sheet_name, self.current_worksheet)
        records = [[x.value for x in row] for row in ws.iter_rows(min_row=start_row)]

        int_columns = int_columns if int_columns else []
        date_columns = date_columns if date_columns else []

        # По каждой строчке
        for r in records:
            # Сцепляем заголовки (английские) из настроек с данными из Excel-файла
            tmp_row = dict(zip(columns, r))

            # Костыль от преобразований данных экселем после ручных правок
            for c in columns:
                # Костыль для столбцов, которые должны быть int
                if tmp_row.get(c) is not None:
                    if c in int_columns and not isinstance(tmp_row.get(c), int):
                        tmp_row[c] = int(tmp_row[c])
                    # Костыль для столбцов, которые должны быть datetime
                    if c in date_columns and not isinstance(tmp_row.get(c), datetime):
                        try:
                            tmp_row[c] = main.parse_date(tmp_row[c])
                        except TypeError:
                            # Костыль, если дата в экселе не текст, а число
                            tmp_row[c] = datetime(1899, 12, 31) + timedelta(days=tmp_row[c] - 1)
                        except ValueError:
                            tmp_row[c] = None
                    # Костыль для столбцов, которые должны быть str
                    if c not in date_columns and c not in int_columns and not isinstance(tmp_row.get(c), str):
                        tmp_row[c] = str(tmp_row.get(c))
            excel_arr.append(tmp_row)

        return excel_arr


def create_excel(list_: List[Union[Dict, List]], fname: str, header: Optional[List] = None,
                 column_width: Optional[List] = None, worksheet: str = 'Sheet', start_row: int = 0, borders=False):
    """
    Создает Excel-файл в формате xlsx
    :param list_: Входной список
    :param fname: Имя файла с путем
    :param column_width: Список с шириной столбцов
    :param worksheet: Имя вкладки
    :param start_row: С какой строки начинать писать
    :return:
    """
    # Если указанной папки не существует - создаем ее (включая родительские, если нужно)
    fd, fn = os.path.split(fname)
    if fd:
        main.check_folder_exist(fd)
    # Создаем книгу
    if fname.split('.')[-1] == 'xlsx':
        create_xlsx(fname, list_, column_width, header, worksheet, start_row, borders)
    else:
        create_xls(fname, list_, column_width, header, worksheet, start_row, borders)


def create_xls(filename: str, data: List[Union[Dict, List]], column_width: Optional[List], header: Optional[List],
               sheet_name: str = 'Sheet', start_row: int = 0, add_borders: bool = False):
    if not filename.endswith('xls'):
        filename += '.xls'
    try:
        wb = xlwt.Workbook()
    except ImportError:
        raise ImportError("xlwt isn't installed. Run: pip install -U xlwt")

    ws = wb.add_sheet(sheet_name)

    common_style = xlwt.XFStyle()
    font = xlwt.Font()
    common_style.alignment.wrap = 1
    if add_borders:
        common_borders = xlwt.Borders()
        common_borders.bottom = xlwt.Borders.THIN
        common_borders.top = xlwt.Borders.THIN
        common_borders.left = xlwt.Borders.THIN
        common_borders.right = xlwt.Borders.THIN

        common_style.borders = common_borders
    common_style.font = font

    if header:
        # Устанавливаем стиль для заголовков
        header_style = xlwt.XFStyle()
        header_style.alignment.wrap = 1
        header_font = xlwt.Font()
        header_font.bold = True

        header_borders = xlwt.Borders()
        if add_borders:
            header_borders.top = xlwt.Borders.THIN
            header_borders.left = xlwt.Borders.THIN
            header_borders.right = xlwt.Borders.THIN
        header_borders.bottom = xlwt.Borders.THIN
        header_style.borders = header_borders
        header_style.font = header_font

        # Заполняем шапку
        for c_index, c_value in enumerate(header):
            ws.write(0, c_index, c_value, header_style)
        start_row += 1

    if column_width:
        # Устанавливаем ширину столбцов
        for c_index, c_width in enumerate(column_width):
            ws.col(c_index).width = 256 * c_width

    # Вставляем значения
    for r_index, row in enumerate(data):
        try:
            columns = list(row.values())
        except AttributeError:
            columns = row
        for c_index, c_value in enumerate(columns):
            # Если значение типа datetime - преобразовываем в строку вида dd.MM.yyyy HH:mm:ss
            if isinstance(c_value, datetime):
                common_style.num_format_str = 'dd.mm.yyyy hh:mm:ss'
            # Если значение типа date - преобразовываем в строку вида dd.MM.yyyy
            elif isinstance(c_value, date):
                common_style.num_format_str = 'dd.mm.yyyy'
            elif isinstance(c_value, int):
                common_style.num_format_str = 'General'
            elif isinstance(c_value, float):
                common_style.num_format_str = 'General'
            else:
                common_style.num_format_str = 'General'

            ws.write(start_row + r_index, c_index, c_value, common_style)
    wb.save(filename)


@deprecated(deprecated_in='2.6.0', current_version=version, details='Use class ExcelX instead')
def create_xlsx(filename: str, data: List[Union[Dict, List]], column_width: Optional[List], header: Optional[List],
                sheet_name: str = 'Sheet', start_row: int = 0, add_borders: bool = False):
    if not filename.endswith('xlsx'):
        filename += '.xlsx'

    try:
        wb = xlsxwriter.Workbook(filename)
    except ImportError:
        raise ImportError("xlsxwriter isn't installed. Run: pip install -U xlsxwriter")

    ws = wb.add_worksheet(sheet_name)

    common_style = wb.add_format({'text_wrap': True})
    if add_borders:
        common_style.set_border()

    if header:
        header_style = wb.add_format({'bold': 1, 'text_wrap': True})
        if add_borders:
            header_style.set_border()
        else:
            header_style.set_bottom()
        # Заполняем шапку
        for c_index, c_value in enumerate(header):
            ws.write(0, c_index, c_value, header_style)
        start_row += 1

    if column_width:
        for c_index, c_width in enumerate(column_width):
            ws.set_column(c_index, c_index, c_width)

    # Вставляем значения
    for r_index, row in enumerate(data):
        try:
            columns = list(row.values())
        except AttributeError:
            columns = row
        for c_index, c_value in enumerate(columns):
            # Если значение типа datetime - преобразовываем в строку вида dd.MM.yyyy HH:mm:ss
            if isinstance(c_value, datetime):
                common_style.num_format = 'dd.mm.yyyy hh:mm:ss'
            # Если значение типа date - преобразовываем в строку вида dd.MM.yyyy
            elif isinstance(c_value, date):
                common_style.num_format = 'dd.mm.yyyy'
            elif isinstance(c_value, int):
                common_style.num_format = 'General'
            elif isinstance(c_value, float):
                common_style.num_format = 'General'
            else:
                common_style.num_format = 'General'

            ws.write(r_index + start_row, c_index, c_value, common_style)
    wb.close()


def read_excel(filename: str, excel_header: Union[list, dict], sheet_name: str = None,
               int_columns: list = None, date_columns: list = None, start_row: int = 1):
    # Считываем Excel-файл
    excel_arr = []
    if isinstance(excel_header, list):
        columns = excel_header
    else:
        columns = [x[1] for x in list(excel_header.values())]

    if filename.endswith('.xlsx'):
        wb = openpyxl.load_workbook(filename)
        ws = wb[sheet_name] if sheet_name else wb.active
        records = [[x.value for x in row] for row in ws.iter_rows(min_row=start_row)]
    else:
        try:
            wb = xlrd.open_workbook(filename)
        except ImportError:
            raise ImportError("xlrd isn't installed. Run: pip install -U xlrd")

        ws = wb.sheet_by_name(sheet_name) if sheet_name else wb.sheet_by_index(0)
        records = [ws.row_values(i) for i in range(start_row - 1, ws.nrows)]

    int_columns = int_columns if int_columns else []
    date_columns = date_columns if date_columns else []

    # По каждой строчке
    for r in records:
        # Сцепляем заголовки (английские) из настроек с данными из Excel-файла
        tmp_row = dict(zip(columns, r))

        # Костыль от преобразований данных экселем после ручных правок
        for c in columns:
            # Костыль для столбцов, которые должны быть int
            if tmp_row.get(c) is not None:
                if c in int_columns and not isinstance(tmp_row.get(c), int):
                    tmp_row[c] = int(tmp_row[c])
                # Костыль для столбцов, которые должны быть datetime
                if c in date_columns and not isinstance(tmp_row.get(c), datetime):
                    try:
                        tmp_row[c] = main.parse_date(tmp_row[c])
                    except TypeError:
                        # Костыль, если дата в экселе не текст, а число
                        tmp_row[c] = datetime(1899, 12, 31) + timedelta(days=tmp_row[c] - 1)
                    except ValueError:
                        tmp_row[c] = None
                # Костыль для столбцов, которые должны быть str
                if c not in date_columns and c not in int_columns and not isinstance(tmp_row.get(c), str):
                    tmp_row[c] = str(tmp_row.get(c))
        excel_arr.append(tmp_row)

    return excel_arr


if __name__ == '__main__':
    excel = ExcelX()
    excel.create_simple_excel('test.xlsx', [{'test': 'test'}], worksheet='AIMS', rewrite=True)
    excel.create_simple_excel('test.xlsx', [{'test': 'test2'}], worksheet='AIMS2', rewrite=False)
